/*
 *  C++ core and functions for BASIN.
 *  Copyright (C) 2006  BASIN team
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
/*! \file Geometry.h
  \brief Contains the Geometry class.

  Geometry defines the dimensionality, metric, boundary condition, and space type of all Region and Data objects.
*/

#ifndef BASIN_GEOMETRY_H
#define BASIN_GEOMETRY_H

#include "basin_base.h"

/** \class Geometry Geometry.h "Geometry.h"
    \brief Geometry is the common property for Data and Region objects.
    
    Every Region, Grid, and particle List has a Geometry associated
 *  with it.  The Geometry includes several key issues:
 *
 *  1. A dimensionality. 
 *
 *  2. A metric. 
 *
 *  3. A boundary condition (periodic or isolated). 
 *
 *  4. A space type (real, fourier)
 * 
 *  By default all top level \link Grid Grids\endlink, which is to say, all \link Grid Grids\endlink which are spawned
 * from a \link Region "Region"\endlink
 *  take the same Geometry as the Region, itself.
 *  All \link Region Regions\endlink within a dataset have the same Geometry but not
 *  necessarily the same topology (boundary conditions).  This will be
 *  set up as we set up the tree structure.
*/
    

class Geometry {
 public:
  /** @name Geometry constructors
      Functions to create Geometry objects.*/
  //@{
  //! Empty Geometry constructor
  Geometry();
  //! Geometry destructor
  ~Geometry();
  //! Explicitly define your Geometry.
  /*! A Geometry is defined by the the dimensionality (an int) the
   * metric type (cartesian, spherical, etc), the boundary type
   * (periodic, or cartesian), the space type (real, fourier, etc.) and the limits of the
   * coordinates...
   */
  Geometry(int nDimsNew,int metricTypeNew, int boundaryTypeNew,int spaceTypeNew,const vector<double>& llNew,const vector<double>& urNew);
  //! Create a copy of an existing geometry
  Geometry(const Geometry&);
  //@}
  /** @name Accessor Functions
      Functions to get members of the Geometry object. */
  //@{
  //! Return the number of dimensions in the current geometry
  int get_ndims();
  //! Returns the metricType
  int get_metrictype(){return metricType;}
  //! Returns the boundaryType
  int get_boundarytype(){return boundaryType;}
  //! Returns the spaceType
  int get_spacetype() {return spaceType;}
  //! Get the type of boundary from the given geometry
  /*!  returns the ll bound if the argument is 0, returns the upper
   * right bound if the argument is 1
   */
  vector<double> get_bound(int iSide);
  //@}
  /** @name Modifier Functions
      Functions to change the value of members.*/
  //@{
  //! Change the number of dimensions
  void set_ndims(int nDimsNew) {nDims=nDimsNew;}
  //! Set the type of metric, eg Cartesian, Spherical, etc.
  int set_metrictype(int metricTypeNew){metricType=metricTypeNew; return 0;}
  //! Set the type of boundary, eg Periodic, Isolated, etc.
  int set_boundarytype(int boundaryTypeNew){boundaryType=boundaryTypeNew; return 0;}
  int set_spacetype(int spaceTypeNew){spaceType=spaceTypeNew; return 0;}
  //@}

  /** @name Distance Calculations
      Calculations based upon the given Geometry attributes.*/
  //@{
  //! Return the distance^2 between two vectors
  float calculate_distance_squared(const vector<double>& x0,const vector<double>& x1);
  //@}
  
      
 protected:
  int nDims; ///< Dimensionality
  ///Metric type
  /** 0     Cartesian\n
      1     Spherical\n
      2     Cylindrical\n
      -1    Other*/
  int metricType;
  ///Boundary type
  /** 0     Isolated\n
      1     Periodic*/
  int boundaryType;
  ///Space type
  /** 0     Real\n
      1     Fourier Hermite\n
      2     Fourier General*/
  int spaceType;
  vector<double> ll; ///< Lower left corner
  vector<double> ur; ///< Upper right corner
};
#endif

