"""
The front end for the basin module. This code will import the C++ interface
and prepare it for use in the Python interpreter.
"""

# import the C++ interface
import _basin

import basin_aux as aux

# other imports
import sys
import types

# in order to ensure proper synchronization
# every basin function must have an MPI_Barrier
# inserted after it. This seems to be necessary
# for one-sided communications to work properly
# when wrapped in the IPython engine.
# The code here will implement
# decorators that will add this call to Barrier on a Function
# or a Class.

def barrierFunctionDecorator(f):
    """ Decorates a free standing function, making sure it ends with a 
    call to an MPI_Barrier."""
    def newF(*args, **kw):
        result = f(*args, **kw)
        _basin.basin_mpi_barrier()
        return result
    newF.__name__ = f.__name__
    newF.__doc__  = f.__doc__
    return newF


def barrierClassDecorator(cls):
    """Decorates all the methods in a class.
    Should return a class that is identical to the original
    except with decorated methods"""

    for name,v in cls.__dict__.items():
        m = getattr(cls, name)
        if type(m) == types.MethodType:
            newM = barrierFunctionDecorator(m)
            setattr(cls, name, newM)

# Used to implement newer style iterators
# Still experimental
def __iter__Func(self):
    self.itr_value = 0
    self.itr_stop = self.size()
    return self

def next_Func(self):
    self.itr_value += 1
    if self.itr_value == self.itr_stop:
        raise StopIteration
    return self.__getitem__(self.itr_value)

def iteratorDecorator(cls):
    setattr(cls,"__iter__",__iter__Func)
    setattr(cls,"next", next_Func)


apply_decorator = {('Attribute',):(iteratorDecorator,)}


# Strings representing the types of Boost Python classes
# and functions
boost_class_type = "<type 'Boost.Python.class'>"
boost_function_type = "<type 'Boost.Python.function'>"

mod = sys.modules[__name__]

basin_funcs = _basin.__dict__.keys()

for name,f in _basin.__dict__.items():
    # test to see if f is really a function
    if str(type(f)) == boost_function_type:
        # get a decorated function
        newF = barrierFunctionDecorator(f)
        # add this function to the namespace of this module
        setattr(mod,name,newF)
        #mod.__dict__[name] = newF
    elif str(type(f)) == boost_class_type:
        # get a class with decorated methods
        barrierClassDecorator(f)
        # apply decorators for specific classes to those
        # classes
        # Experimental, so commented out for now
        #for class_list,dec_list in apply_decorator.items():
        #    if name in class_list:
        #        for dec in dec_list:
        #            dec(f)
        setattr(mod,name,f)
        #mod.__dict__[name] = newCls



