#! /bin/bash

function print_usage 
{
    echo "This script will start the BASIN remote environment. It should 
be run on your cluster. The controller will be launched on this machine
while the engines will be started using mpiexec. You should have your 
MPI environment created before using this command.

Arguments
    -n --nodes  number of nodes to bring up. This argument is required

    -p --port   the base number port to use. It will also use the next two 
                higher ports from the base listed. For example, if the argument
                is 1000 the ports 1000, 1001, and 1002 will be used by the 
                system. This argument is optional.
    -h --hostname   the name of the host where the controller is running. 
                    This argument is optional. If it is blank, the hostname
                    program will be used to find the value. This argument
                    is useful when the machines hostname is not an valid 
                    name on the network."

}

num_nodes=0
port_base=10105
eng_port=
task_port=
hostname=`hostname`
visit_dir=/usr/local/apps/visit/current/linux-x86_64

while [ "$1" != "" ]; do
    case $1 in
        -n | --nodes )
            shift
            num_nodes=$1
            ;;
        -p | --port )
            shift
            port_base=$1
            ;;
        -h | --hostname )
            shift
            hostname=$1
            ;;
        * )
            print_usage
            exit
            ;;
    esac
    shift
done

if [ $num_nodes -eq 0 ]; then
    print_usage
    exit
fi

# Check if we can source the environment variable settings
if [ -f basin_remote_env.sh ]; then
    source basin_remote_env.sh
    env_found="basin_remote_env.sh"
else
    env_found="No environment file found."
fi


# Create file containing path to client scripts - James
if ! [ -d $HOME/.basin ]; then
	mkdir $HOME/.basin
fi
if ! [ -d $HOME/.basin/host ]; then
        mkdir $HOME/.basin/host
fi
script_path=$0
if [ -L $script_path ] ; then
	script_path=`ls -l $script_path | gawk '{ print $11 }'`
else
	script_path=`pwd`/$script_path
fi
script_path=`dirname $script_path`
echo "== Basin Client Paths =="
echo "Scripts Path: "$script_path
echo "  VisIt Path: "`which visit`
echo "========================"
echo $script_path > $HOME/.basin/host/client_script.path
echo "`which visit`" > $HOME/.basin/host/visit.path
echo "========================"
echo "Using environment settings from file: $env_found"
echo 


# If visit_dir has been set, then set the environment vars accordingly
if [ -n "$visit_dir" ]; then
    LD_LIBRARY_PATH="$visit_dir/lib:$LD_LIBRARY_PATH"
    VISITPLUGINDIR="$visit_dir/plugins"
fi

# Set the environment for the BASIN kernel libraries
if [ "$BASIN_LIB_DIR" ]; then
    LD_LIBRARY_PATH="$BASIN_LIB_DIR:$LD_LIBRARY_PATH"
fi

# Make sure all the necessary environment variables are sent to the cluster
env_args="-env PYTHONPATH /home/djones/svn/basin-py2.5/basin_remote/trunk/usr/lib64/python2.5/site-packages:$PYTHONPATH"

# LD_LIBRARY_PATH and VISITPLUGINDIR are needed for visit support. These may 
# not already be defined, so test for existence before using them
if [ "$LD_LIBRARY_PATH" ]; then
    env_args=$env_args" -env LD_LIBRARY_PATH $LD_LIBRARY_PATH"
fi

if [ "$VISITPLUGINDIR" ]; then
    env_args=$env_args" -env VISITPLUGINDIR $VISITPLUGINDIR"
fi

if [ $port_base -ne 10105 ]; then
    # Launch with customized ports
    
    let eng_port=$port_base+1
    let task_port=$port_base+2
    
    ipcontroller --remote-cont-port=$port_base --engine-port=$eng_port --task-port=$task_port &
    controller_pid=$!
    sleep 3
    
    # Test that the controller started ok
    if [ "x"`jobs | awk '{print $2}'` = 'xRunning' ]; then
        mpiexec -n $num_nodes $env_args ipengine --controller-ip=$hostname --controller-port=$eng_port &
        engine_pid=$!
    else
        echo "Error launching the controller, exiting."
        exit
    fi

else
    # Launch with default ports

    ipcontroller &
    controller_pid=$!
    sleep 3

    # Test that the controller started ok
    if [ "x"`jobs | awk '{print $2}'` = 'xRunning' ]; then
        mpiexec -n $num_nodes $env_args ipengine --controller-ip=$hostname &
        engine_pid=$!
    else
        echo "Error launching the controller, exiting."
        exit
    fi

fi
 
# create a sig handler that will kill our two children
trap "{ kill -9 $controller_pid; kill -9 $engine_pid; exit; }" SIGINT SIGKILL

# make sure the script doesn't exit before its children do 
wait
    

